package be.iminds.ilabt.jfed.ui.cli2.instruction;

import be.iminds.ilabt.jfed.experiment.setup.*;
import be.iminds.ilabt.jfed.experiment.setup.config.*;
import com.fasterxml.jackson.annotation.JsonCreator;
import com.fasterxml.jackson.annotation.JsonIgnore;
import com.fasterxml.jackson.annotation.JsonProperty;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import java.util.Collections;
import java.util.List;

/**
 * RunInstruction: Instructions for a "run":
 *   - creating a slice (optional)
 *   - generating/reading an ESpec/RSpec
 *   - provisioning an ESpec/RSpec
 *   - waiting until ready (sometimes optional)
 *   - delete when something goes wrong
 */
public class RunInstruction extends Instruction {
    public static class SshKeySource {
        private final boolean userCert;
        private final boolean userKeys;
        private final boolean shareWith;
        private final boolean rspec;
        @Nonnull private final List<String> extraUsernames; //username part of URN  (will lookup keys at authority)
        @Nonnull private final List<String> extraKeys; //direct ssh keys in openssh format

        @JsonCreator
        public SshKeySource(@JsonProperty("userCert") @Nullable Boolean userCert,
                            @JsonProperty("userKeys") @Nullable Boolean userKeys,
                            @JsonProperty("shareWith") @Nullable Boolean shareWith,
                            @JsonProperty("rspec") @Nullable Boolean rspec,
                            @JsonProperty("extraUsernames") @Nullable List<String> extraUsernames,
                            @JsonProperty("extraKeys") @Nullable List<String> extraKeys) {
            this.userCert = userCert == null ? new SshKeySource().userCert : userCert;
            this.userKeys = userKeys == null ? new SshKeySource().userKeys : userKeys;
            this.shareWith = shareWith == null ? new SshKeySource().shareWith : shareWith;
            this.rspec = rspec == null ? new SshKeySource().rspec : rspec;
            this.extraUsernames = extraUsernames == null ? new SshKeySource().extraUsernames : extraUsernames;
            this.extraKeys = extraKeys == null ? new SshKeySource().extraKeys : extraKeys;
        }

        public SshKeySource() {
            this.userCert = true;
            this.userKeys = true;
            this.shareWith = true;
            this.rspec = true;
            this.extraUsernames = Collections.emptyList();
            this.extraKeys = Collections.emptyList();
        }

        @JsonProperty
        public boolean isUserCert() {
            return userCert;
        }

        @JsonProperty
        public boolean isUserKeys() {
            return userKeys;
        }

        @JsonProperty
        public boolean isShareWith() {
            return shareWith;
        }

        @JsonProperty
        public boolean isRspec() {
            return rspec;
        }

        @JsonProperty
        @Nonnull
        public List<String> getExtraUsernames() {
            return extraUsernames;
        }

        @JsonProperty
        @Nonnull
        public List<String> getExtraKeys() {
            return extraKeys;
        }
    }

    public static class ShareWith {
        @Nonnull private final List<String> users; //urns or usernames
        private final boolean projectMembers;

        @JsonCreator
        public ShareWith(@JsonProperty("users") @Nonnull List<String> users,
                         @JsonProperty("projectMembers") boolean projectMembers) {
            this.users = users;
            this.projectMembers = projectMembers;
        }

        public ShareWith() {
            this.users = Collections.emptyList();
            this.projectMembers = false;
        }

        @JsonProperty
        @Nonnull
        public List<String> getUsers() {
            return users;
        }

        @JsonProperty
        public boolean isProjectMembers() {
            return projectMembers;
        }
    }

    public static class DeleteOn {
        private final boolean failCreate;           //t
        private final boolean failBecomeReady;      //f
        private final boolean failConnectivityTest; //f
        private final boolean failLinkTest;         //f
        private final boolean failESpec;            //f   //espec and other post-ready steps (such as wait-for, old executes, old ansible, ...)

        @JsonCreator
        public DeleteOn(@JsonProperty("failCreate") @Nullable Boolean failCreate,
                        @JsonProperty("failBecomeReady") @Nullable Boolean failBecomeReady,
                        @JsonProperty("failConnectivityTest") @Nullable Boolean failConnectivityTest,
                        @JsonProperty("failLinkTest") @Nullable Boolean failLinkTest,
                        @JsonProperty("failESpec") @Nullable Boolean failESpec) {
            this.failCreate = failCreate == null ? new DeleteOn().failCreate : failCreate;
            this.failBecomeReady = failBecomeReady == null ? new DeleteOn().failBecomeReady : failBecomeReady;
            this.failConnectivityTest = failConnectivityTest == null ? new DeleteOn().failConnectivityTest : failConnectivityTest;
            this.failLinkTest = failLinkTest == null ? new DeleteOn().failLinkTest : failLinkTest;
            this.failESpec = failESpec == null ? new DeleteOn().failESpec : failESpec;
        }

        public DeleteOn() {
            this.failCreate = true;
            this.failBecomeReady = false;
            this.failConnectivityTest = false;
            this.failLinkTest = false;
            this.failESpec = false;
        }

        @JsonProperty
        public boolean isFailCreate() {
            return failCreate;
        }

        @JsonProperty
        public boolean isFailBecomeReady() {
            return failBecomeReady;
        }

        @JsonProperty
        public boolean isFailConnectivityTest() {
            return failConnectivityTest;
        }

        @JsonProperty
        public boolean isFailLinkTest() {
            return failLinkTest;
        }

        @JsonProperty
        public boolean isFailESpec() {
            return failESpec;
        }
    }


    @Nonnull private final ExperimentSetupConfig experiment;

    /**
     * list of usernames + optionally the special value PROJECT
     *
     * (this shares the slice, the keys are added according to sshKeys)
     */
    @Nonnull private final ShareWith shareWith;

    @Nonnull private final SshKeySource sshKeys;

//    /**
//     * String that allows a lot of formats:
//     *   - 2017-06-05T04:05:06Z    (RFC3339)
//     *   - 2017-06-05 04:05:06Z    (ISO 8601)
//     *   - in 5 days
//     *   - in 30 minutes
//     *   - 10 hours
//     *   - 10h
//     *   - 45 min
//     *  NOT - 10                    (numbers are interpreted as hours)  (BAD idea)
//     */
//    @Nonnull private final String expiration;

    @Nonnull private final DeleteOn deleteOn;


    /**
     *
     * @param action
     * @param user
     * @param actionOutputs
     * @param debugOutputs
     * @param callOutputs
     * @param experiment
     * @param shareWith who to share the slice with
     * @param sshKeys
     * @param deleteOn
     */
    @JsonCreator
    public RunInstruction(@JsonProperty("action") @Nonnull Action action,
                          @JsonProperty("user") @Nullable User user,
                          @JsonProperty("actionOutputs") @Nullable List<ActionOutput> actionOutputs,
                          @JsonProperty("debugOutputs") @Nullable List<DebugOutput> debugOutputs,
                          @JsonProperty("callOutputs") @Nullable List<CallOutput> callOutputs,
                          @JsonProperty("proxy") @Nullable ProxySettings proxy,
                          @JsonProperty("experiment") @Nullable ExperimentSetupConfig experiment,
                          @JsonProperty("shareWith") @Nullable ShareWith shareWith,
                          @JsonProperty("sshKeys") @Nullable SshKeySource sshKeys,
                          @JsonProperty("deleteOn") @Nullable DeleteOn deleteOn) {
        super(action, user, actionOutputs, debugOutputs, callOutputs, proxy);
        this.experiment = experiment == null ? new RunInstruction(false).experiment : experiment;
        this.shareWith = shareWith == null ? new RunInstruction(false).shareWith : shareWith;
        this.sshKeys = sshKeys == null ? new RunInstruction(false).sshKeys : sshKeys;
        this.deleteOn = deleteOn == null ? new RunInstruction(false).deleteOn : deleteOn;
    }

    private RunInstruction(boolean demo) {
        super(Action.RUN, demo);
        Slice slice = new Slice("myExp1", false, false, 120l, true,
                Slice.ProjectSource.PROVIDED, true, "myProject");
        WaitForReady waitForReady = new WaitForReady(30l, null, true);
        Provision provision = new Provision(null, true, null);
        this.experiment = demo ?
                new ExperimentSetupConfig(slice, new ProvidedContentRequestRSpec(
                        RequestRSpec.RequestRSpecSource.PROVIDE_CONTENT,
                        "<rspec type=\"request\">\n   <node>\n      <!--my_rspec-->\n   </node>\n</rspec>\n",
                        false,
                        "urn:publicid:IDN+example.com+authority+am",
                        false,
                        false
                ),
                        null, provision, waitForReady, null, true) :
                new ExperimentSetupConfig();
        this.shareWith = new ShareWith();
        this.sshKeys = new SshKeySource();
        this.deleteOn = new DeleteOn();
    }

    @JsonProperty
    @Nonnull
    public ExperimentSetupConfig getExperiment() {
        return experiment;
    }

    @JsonProperty
    @Nonnull
    public ShareWith getShareWith() {
        return shareWith;
    }

    @JsonProperty
    @Nonnull
    public SshKeySource getSshKeys() {
        return sshKeys;
    }


    @JsonProperty
    @Nonnull
    public DeleteOn getDeleteOn() {
        return deleteOn;
    }


    @JsonIgnore
    @Nonnull
    public Slice getSlice() {
        return experiment.getSlice();
    }

    @JsonIgnore
    @Nonnull
    @Override
    public Instruction.Action getActionOfClass() {
        return Action.RUN;
    }
}
